/*
 * GETCNFG.C - Configuration file handler.
 *
 * A configuration file is a text file of form:
 *
 *          configuration_file   =  { configuration_parm }.
 *          configuration_parm   =  parm_name '=' parm_value.
 *          parm_name            =  string. (must be a single token)
 *          parm_value           =  string. (can't extend past end-of-line)
 *
 * In other words, a configuration file consists of parameter name/value
 * pairs, where name and value are separated by an equal sign (and any
 * amount of whitespace.)  The parameter name cannot contain imbedded
 * whitespace.  The parameter value cannot extend beyond the end of the 
 * line in which it begins.
 *
 * The user defines a table of PARM's (see getcnfg.h), each entry of which
 * specifies:
 *
 *       -  a parameter name to be recognized
 *       -  the expected type of the parameter's values
 *          (BOOLEAN_P, CHARACTER_P, INTEGER_P, LONG_P, DOUBLE_P, STRING_P,
 *          or FUNCTION_P)
 *       -  a pointer to a variable to be assigned the parameter's value
 *          (for types BOOLEAN_P, CHARACTER_P, INTEGER_P, LONG_P, DOUBLE_P and
 *          STRING_P), or a pointer to a function of one character pointer,
 *          returning an integer (for type FUNCTION_P)
 *       -  the size of the variable to be assigned the parameter's value
 *       -  a pointer to an informative message related to the parameter
 *
 * The PARM table specifies how values in a configuration file will be
 * interpreted to set run-time parameters.  Currently, no type validation
 * is performed.
 *
 * For BOOLEAN_P parameters, the value supplied should be "TRUE" or "YES"
 * or "ON" or "ACTIVE" to set the corresponding boolean variable "on";
 * other values will set the corresponding variable "off".
 *
 * Note that STRING_P variable pointers can be either the address of the
 * first char of a char array or the address of a char pointer.  The two
 * cases are distinguished by coding a 0 as the variable size for a pointer
 * to a char pointer, and coding the size of the array otherwise.
 *
 * $Header:   C:/fe/shared/vcs/getcnfg.c_v   1.5   24 Jan 1990 09:54:54   rji  $
 *
 * $Log:   C:/fe/shared/vcs/getcnfg.c_v  $
 * 
 *    Rev 1.5   24 Jan 1990 09:54:54   rji
 * Added LONG_P parameter processing for long config parms.  (KGW)
 * Comment and cosmetic update.  (RJI)
 * 
 *    Rev 1.4   25 Dec 1989 03:36:46   rji
 * Brought header block comment up-to-date.
 * 
 *    Rev 1.3   05 Nov 1989 21:40:58   rji
 * Now allows multi-token values (which cannot extend past the end-of-line,
 * however).  Values strings are trimmed of trailing spaces (they are already
 * trimmed of leading spaces).
 * 
 *    Rev 1.2   13 May 1989  2:07:08   rji
 * Improved display function print_cnfg_usage(), preventing use of null pointers
 * for variable length strings.  Test driver now always prints out usage.
 * Split error return value into -1 if file not found, -2 if parameter 
 * specification error.
 * 
 * 
 *    Rev 1.1   22 Feb 1989 13:28:24   rji
 * Now allows custom-sized STRING_P parms, by specifying address of a char
 * pointer instead of first char of char array and giving a 0 length.  Also,
 * changed "register" to "REGi" and improved comments.
 * 
 *    Rev 1.0   07 Dec 1988 23:10:18   rji
 * Initial revision.
 * 
*/

/*=========================================================================*/
/*=                                                                       =*/
/*=      I N C L U D E S,  D E F I N E S,  D E C L A R A T I O N S        =*/
/*=                                                                       =*/
/*=========================================================================*/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         GETCNFG.C Includes

  $PATHS$
       Functions\C
       Shared\Includes
       Shared\Utils
       Modules\$./FL$

   $3$

   See Also:
        $/SEE(COMMON.H Macros)$
        $/SEE(GETCNFG.H Macros)$

   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* INCLUDES                                                                */
/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <common.h>
#include <dalloc.h>
#include <getcnfg.h>
/*                                                                   $end$ */

#define  MAXPARMNAMESZ     80
#define  MAXVALUESZ        80

static PARM    *findparm( char *, PARM *, size_t );
static void    setparm( PARM *, char * );
static void    print_cnfg_usage( PARM *, size_t );

/*=========================================================================*/
/*=                                                                       =*/
/*=                F U N C T I O N   D E F I N I T I O N S                =*/
/*=                                                                       =*/
/*=========================================================================*/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         getcnfg()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* GETCNFG                                                                 */
/*                                                                         */
/* Process specified configuration file.  Returns -1 if configuration file */
/* not found, -2 if error detected in configuration file; otherwise,       */
/* returns number of parameters processed (could be zero.)                 */
/*-------------------------------------------------------------------------*/

int   getcnfg( cnfg_file_name, parm_tbl, parm_tbl_sz )
   char    *cnfg_file_name;            /* in: name of configuration file   */
   PARM    *parm_tbl;                  /* in: parameter description table  */
   size_t   parm_tbl_sz;               /* in: # of entries in parm_tbl     */
/*                                                                   $end$ */
{
   FILE       *fp;                     /* for access to configuration file */
   char        parm[MAXPARMNAMESZ];    /* holds last read parameter name   */
   char        val[MAXVALUESZ];        /* holds last read parameter value  */
   REG1 PARM  *p;                      /* ptr to found parm_tbl entry      */
   REG2 int    t;                      /* # of fscanf conversions made     */
   REG3 int    parm_cnt = 0;           /* # of parameters processed        */


   if ( (fp = fopen( cnfg_file_name, "r" )) == NULL )
      return ( -1 );

   /* initialize parameter/value strings */

   parm[0] = '\0';
   val[0] = '\0';

   /* parameter/value pair processing loop */

   for (;;)
   {
      if ( (t = fscanf( fp, " %[^ \t\n=] = %[^\t\n=]", parm, val )) == EOF )
         break;

      if ( t == 2 )
      {
         if ( p = findparm( parm, parm_tbl, parm_tbl_sz ) )
         {
            setparm( p, val );
            ++parm_cnt;
         }
         else
         {
            fprintf( stderr, "Unrecognized Parameter \"%s\".  "
                             "Valid Parameters are:\n\n", parm );
            print_cnfg_usage( parm_tbl, parm_tbl_sz );
            parm_cnt = -2 ;
            break;
         }
      }
      else
      {
            fprintf( stderr, "Ill-formed Parameter Specification: "
                             "\"%s\" = \"%s\".\n"
                             "Check Parameter Values below:\n\n", parm, val );
            print_cnfg_usage( parm_tbl, parm_tbl_sz );
            parm_cnt = -2 ;
            break;
      }
   }

   fclose( fp );

   return ( parm_cnt );

} /* getcnfg() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         findparm()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* FINDPARM                                                                */
/*                                                                         */
/* Search parameter table for entry corresponding to specified parameter   */
/* name.  Returns pointer to matched table entry, if found; otherwise,     */
/* returns NULL.                                                           */
/*-------------------------------------------------------------------------*/

static PARM    *findparm( name, parm_tbl, parm_tbl_sz )
   REG1 char     *name;                /* in: parameter name to look up    */
   REG2 PARM     *parm_tbl;            /* in: ptr to 1st parm table entry  */
   REG3 size_t    parm_tbl_sz;         /* in: # of entries in parm_tbl     */
/*                                                                   $end$ */
{
   for ( ; parm_tbl_sz-- > 0 ; ++parm_tbl )
      if ( !stricmp( name, parm_tbl->name ) )
         return ( parm_tbl );

   return ( NULL );

} /* findparm() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         setparm()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* SETPARM                                                                 */
/*                                                                         */
/* Set parameter value according to matched PARM table entry specs.        */
/*-------------------------------------------------------------------------*/

static void setparm( parm_tbl_entry, val )
   REG1 PARM     *parm_tbl_entry;      /* in: ptr to parm table entry      */
   REG2 char     *val;                 /* in: parm value (as string)       */
/*                                                                   $end$ */
{
   int   lastch;                       /* index to last character of val   */

   /* first, trim trailing blanks (leading blanks already trimmed) */

   lastch = strlen( val ) - 1;
   while ( isspace( val[lastch] ) )
      --lastch;
   val[lastch+1] = '\0';

   /* now, set parameter value according to type */

   switch ( parm_tbl_entry->type )
   {
   case BOOLEAN_P:
      if (   !stricmp( val, "TRUE" ) || !stricmp( val, "YES" )
          || !stricmp( val, "ON" )   || !stricmp( val, "ACTIVE" ) )
         *(int *)parm_tbl_entry->var = 1;
      else
         *(int *)parm_tbl_entry->var = 0;
      break;
   case CHARACTER_P:
      *(char *)parm_tbl_entry->var = *val;
      break;
   case INTEGER_P:
      *(int *)parm_tbl_entry->var = atoi( val );
      break;
   case LONG_P:
      *(long *)parm_tbl_entry->var = atol( val );
      break;
   case DOUBLE_P:
      *(double *)parm_tbl_entry->var = atof( val );
      break;
   case STRING_P:
      if ( parm_tbl_entry->varsize == 0 )
      {
         /* parm var is a pointer, so obtain storage before copying value */
         *(char **)parm_tbl_entry->var = d_malloc( strlen( val ) + 1 );
         if ( *(char **)parm_tbl_entry->var == NULL )
            fprintf( stderr, "setparm: Cannot Allocate String Space: "
                             "%u bytes requested\n", strlen( val ) + 1 );
         else
            strcpy( *(char **)parm_tbl_entry->var, val );
      }
      else
         /* parm var is an array, so need only copy value */
         strscpy( (char *)parm_tbl_entry->var, val, parm_tbl_entry->varsize );
      break;
   case FUNCTION_P:
      /* the following non-portable code thanks to Intel architecture */
#if (defined(M_I86CM) || defined(M_I86LM) || defined(M_I86HM))
      (*(int (*)(char *))(long)(parm_tbl_entry->var))( val );
#else
      (*(int (*)(char *))(int)(parm_tbl_entry->var))( val );
#endif
      break;
   default:
      fprintf( stderr, "setparm: Internal Error: Bad Parameter Type: %d\n",
                                                     parm_tbl_entry->type );
      break;
   }

} /* setparm() */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         print_cnfg_usage()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* PRINT_CNFG_USAGE                                                        */
/*                                                                         */
/* Print out recognized parameter names, usage info and current values.    */
/*-------------------------------------------------------------------------*/

static void print_cnfg_usage( parm_tbl, parm_tbl_sz )
   REG1 PARM     *parm_tbl;            /* in: ptr to 1st parm table entry  */
   REG2 size_t    parm_tbl_sz;         /* in: # of entries in parm_tbl     */
/*                                                                   $end$ */
{
   for ( ; parm_tbl_sz-- > 0 ; ++parm_tbl )
   {
      fprintf( stderr, "%-15s : %-40s ",
                 parm_tbl->name, parm_tbl->msg );

      switch ( parm_tbl->type )
      {
      case BOOLEAN_P:
         fprintf( stderr, "(value: %s)\n",
                           *(int *)parm_tbl->var ? "TRUE" : "FALSE" );
         break;
      case CHARACTER_P:
         fprintf( stderr, "(value: '%c')\n", *(char *)parm_tbl->var );
         break;
      case INTEGER_P:
         fprintf( stderr, "(value: %d)\n", *(int *)parm_tbl->var );
         break;
      case LONG_P:
         fprintf( stderr, "(value: %ld)\n", *(long *)parm_tbl->var );
         break;
      case DOUBLE_P:
         fprintf( stderr, "(value: %g)\n", *(double *)parm_tbl->var );
         break;
      case STRING_P:
         if ( parm_tbl->varsize == 0 )
            if ( *(char *) parm_tbl->var != NULL )
               fprintf( stderr, "(value: \"%s\")\n", *(char **) parm_tbl->var );
            else
               fprintf( stderr, "(value: \"\")\n", *(char **) parm_tbl->var );
         else
            fprintf( stderr, "(value: \"%s\")\n", (char *) parm_tbl->var );
         break;
      case FUNCTION_P:
         fprintf( stderr, "\n" );
         break;
      default:
         break;
      }
   }

} /* print_cnfg_usage() */


#if defined(DRIVER)
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         GETCNFG.C main() Test Driver

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*=========================================================================*/
/*=                                                                       =*/
/*=                       T E S T   D R I V E R                           =*/
/*=                                                                       =*/
/*=========================================================================*/

/* function prototypes  */

int   main( int, char ** );
char  *pname( char * );
int   f( char * );

/* PARM specification table   */

double   percent;
long     count;
int      mxatts, underwri_opt;
char     c, histfl[MAXVALUESZ], set_by_f[MAXVALUESZ];
char    *p;

PARM  config[] =
{
   { "UNDERWRITE", BOOLEAN_P, &underwri_opt, sizeof(int), "Boolean - (TRUE/FALSE YES/NO ON/OFF)" },
   { "SPECCHAR", CHARACTER_P, &c, sizeof(char), "Must be a capital letter" },
   { "MXATTS", INTEGER_P, &mxatts, sizeof(int), "Must be between 7 and 255" },
   { "COUNT", LONG_P, &count, sizeof(long), "Must be a long integer" },
   { "PERCENT", DOUBLE_P, &percent, sizeof(double), "Must be a real number" },
   { "HISTORYFILE", STRING_P, histfl, sizeof(histfl), "Must be valid filename" },
   { "PPSTRING", STRING_P, &p, 0, "Must be valid filename" },
   { "FUNCKEY", FUNCTION_P, f, sizeof(void *), "Value is passed to function \"f\"" }
};
/*                                                                   $end$ */

/*-------------------------------------------------------------------------*/

int   f( s )
   char *s;
{
   strcpy( set_by_f, s );
   return ( (int) *s );
}

/*-------------------------------------------------------------------------*/

int   main( argc, argv )
   int      argc;
   char   **argv;
{
   char    *prog;
   int      n;
   size_t   tbl_size = sizeof(config)/sizeof(config[0]);

   prog = pname( argv[0] );

   while ( --argc > 0  &&  (*++argv)[0] == '-' )
      switch ( (*argv)[1] )
      {
         default:
            fprintf( stderr, "usage: %s [<config_file>]\n", prog );
            exit( 1 );
      } /* switch */

   if ( argc > 0 )
      n = get_configuration( *argv, config, tbl_size );
   else
      n = get_configuration( "CON", config, tbl_size );

   if ( n >= 0 )
   {
      fprintf( stderr, "\nSuccessful Configuration Table Parse\n\n" );
      print_cnfg_usage( config, tbl_size );
      fprintf( stderr, "\nValue passed the function \"f\": \"%s\"\n",
                                                             set_by_f );
   }
   else if ( n == -1 )
      fprintf( stderr, "\nCannot find Configuration Table: \"%s\"\n\n", *argv );
   else
      fprintf( stderr, "\nError in Configuration Table: \"%s\"\n\n", *argv );

   return ( 0 );

} /* main() */

#endif
